#include <LiquidCrystal_I2C.h>

// define notes as frequencies
#define NOTE_B0  31
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62
#define NOTE_C2  65
#define NOTE_CS2 69
#define NOTE_D2  73
#define NOTE_DS2 78
#define NOTE_E2  82
#define NOTE_F2  87
#define NOTE_FS2 93
#define NOTE_G2  98
#define NOTE_GS2 104
#define NOTE_A2  110
#define NOTE_AS2 117
#define NOTE_B2  123
#define NOTE_C3  131
#define NOTE_CS3 139
#define NOTE_D3  147
#define NOTE_DS3 156
#define NOTE_E3  165
#define NOTE_F3  175
#define NOTE_FS3 185
#define NOTE_G3  196
#define NOTE_GS3 208
#define NOTE_A3  220
#define NOTE_AS3 233
#define NOTE_B3  247
#define NOTE_C4  262
#define NOTE_CS4 277
#define NOTE_D4  294
#define NOTE_DS4 311
#define NOTE_E4  330
#define NOTE_F4  349
#define NOTE_FS4 370
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define NOTE_C6  1047
#define NOTE_CS6 1109
#define NOTE_D6  1175
#define NOTE_DS6 1245
#define NOTE_E6  1319
#define NOTE_F6  1397
#define NOTE_FS6 1480
#define NOTE_G6  1568
#define NOTE_GS6 1661
#define NOTE_A6  1760
#define NOTE_AS6 1865
#define NOTE_B6  1976
#define NOTE_C7  2093
#define NOTE_CS7 2217
#define NOTE_D7  2349
#define NOTE_DS7 2489
#define NOTE_E7  2637
#define NOTE_F7  2794
#define NOTE_FS7 2960
#define NOTE_G7  3136
#define NOTE_GS7 3322
#define NOTE_A7  3520
#define NOTE_AS7 3729
#define NOTE_B7  3951
#define NOTE_C8  4186
#define NOTE_CS8 4435
#define NOTE_D8  4699
#define NOTE_DS8 4978
#define REST      0

// maximum size allowed for a sequence of sounds
#define MAX_SEQUENCE 40

#define GREEN_LED PD7
#define RED_LED PD6

// change this to make the song slower or faster
int tempo[] = {300, 240, 320, 200, 340, 260};

// change this to whichever pin you want to use
int buzzer = 4;

// notes of the sound followed by the duration

// Snare Drum
int sound1[] = {
  REST, 4,        // Rest, 1 quarter note
  NOTE_D4, 8,     // D4, 1 eighth note
  REST, 8,        // Rest, 1 eighth note
  REST, 4,        // Rest, 1 quarter note
  NOTE_D4, 8,     // D4, 1 eighth note
  REST, 8,        // Rest, 1 eighth note
  REST, 4,        // Rest, 1 quarter note
  NOTE_D4, 8,     // D4, 1 eighth note
  REST, 4,        // Rest, 1 quarter note
};

// Hi-Hat
int sound2[] = {
  REST, 8,        // Rest, 1 eighth note
  NOTE_E4, 8,     // E4, 1 eighth note
  REST, 8,        // Rest, 1 eighth note
  NOTE_E4, 8,     // E4, 1 eighth note
  REST, 8,        // Rest, 1 eighth note
  NOTE_E4, 8,     // E4, 1 eighth note
  REST, 8,        // Rest, 1 eighth note
  NOTE_E4, 8,     // E4, 1 eighth note
};

// Bassline
int sound3[] = {
  NOTE_C3, 4,     // C3, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_C3, 4,     // C3, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_C3, 4,     // C3, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_C3, 4,     // C3, 1 quarter note
};

// Tom Drum
int sound4[] = {
  NOTE_A2, 4,     // A2, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_A2, 4,     // A2, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_A2, 4,     // A2, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_A2, 4,     // A2, 1 quarter note
};

// Synth
int sound5[] = {
  NOTE_E5, 4,     // E5, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_E5, 4,     // E5, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_E5, 4,     // E5, 1 quarter note
  REST, 4,        // Rest, 1 quarter note
  NOTE_E5, 4,     // E5, 1 quarter note
};

// Melody
int sound6[] = {
  NOTE_C4, 4,     // C4, 1 quarter note
  NOTE_D4, 4,     // D4, 1 quarter note
  NOTE_E4, 4,     // E4, 1 quarter note
  NOTE_F4, 4,     // F4, 1 quarter note
  NOTE_E4, 4,     // E4, 1 quarter note
};

// array of all sounds, where soudX is indexed as the Xth element
int *all_sounds[6] = {sound1, sound2, sound3, sound4, sound5, sound6};

// array that will store all sounds played keeping the order
int sequence[MAX_SEQUENCE];
// index to iterate through sequence
int sequence_idx = 0;

// used for printing
char *sounds_name[] = {"Snare Drum", "Hi-Hat", "Bassline", "Tom Drum", "Synth", "Melody"};

// sizeof gives the number of bytes, each int value is composed of two bytes (16 bits)
// there are two values per note (pitch and duration), so for each note there are four bytes
int notes[6];

// this calculates the duration of a whole note in ms
int wholenote[6];

int divider = 0, noteDuration = 0;

///////////////////////

// used for debounce
unsigned int current_time;
unsigned int old_time;

unsigned int pushbuttonsCount = 6; //number of push buttons and resistors
unsigned long resistorValue = 1800; // resistor between each buttons
unsigned long resistorToGround = 220000; // resistor between the set of buttons and the GND

byte inPin = A0; // pin for reading the value of voltage when a button is pressed
float sourceVoltage = 5; // VCC
char *pushbutton_name[] = {"PB1", "PB2", "PB3", "PB4", "PB5", "PB6"}; // used for printing
bool pushbutton_value[] = {0, 0, 0, 0, 0, 0}; // this holds the status of each push button

// used for computing the voltage
int pushDelayTime = 100;
float V_tolerance = 0.05; // 5%
const byte BIT_SOLUTION = 10;

// LCD
LiquidCrystal_I2C lcd(0x27,16,2);

// keeps the state of the Play/Reset button
volatile bool playButtonPressed = false;
volatile bool resetButtonPressed = false;

uint16_t interval; // Initial interval between LED transitions

// keeps the state of the Green/Red LED
bool GREEN_ON = false;
bool RED_ON = false;

void readPushButtons();

// Interrupt service routine for Play Button
ISR(INT1_vect)
{
  // checks if the interrupt is triggered only by the press of the Play button
  // and not by noise or voltage changes
  if (bit_is_clear(PIND, PD3)) {
    playButtonPressed = true;
  }
}

// Interrupt service routine for Reset Button
ISR(INT0_vect)
{
  resetButtonPressed = true;
}

// Interrupt service routine for Timer1 compare match A
ISR(TIMER1_COMPA_vect) {
  interval++;

  if (interval == 36 && GREEN_ON) {
    Serial.println("18 seconds passed, take a break or play the sequence!");

    GREEN_ON = false;
    digitalWrite(GREEN_LED, LOW);   // Turn off the GREEN LED

    RED_ON = true;
    digitalWrite(RED_LED, HIGH);   // Turn on the RED LED

    interval = 0; // reset the interval
  }

  if (interval == 10 && RED_ON) {
    Serial.println("5 seconds passed, you can play again!");
    
    RED_ON = false;
    digitalWrite(RED_LED, LOW);   // Turn on the RED LED
    
    interval = 0; // reset the interval
  }
}

void setup()
{
  Serial.begin(9600);

  pinMode(RED_LED, OUTPUT);
  pinMode(GREEN_LED, OUTPUT);
  pinMode(buzzer, OUTPUT);
  pinMode(inPin, INPUT);

  setup_interrupts();
  setup_sounds();
  setup_timer1();

  lcd.init(); // Start LCD
  lcd.clear();         
  lcd.backlight();      // Make sure backlight is on
  
  // Print a message on both lines of the LCD.
  lcd.setCursor(3,0);
  lcd.print("Music Pad");
  
  lcd.setCursor(2,1);
  lcd.print("Bocanu Raul");

  Serial.println("");
  Serial.println("//////////////////////////////////");
  Serial.println("");

  Serial.println("Welcome to the MusicPad! Let's make some music!");
}

void loop()
{
  // can only play/reset if the green LED is off
  if (playButtonPressed && !GREEN_ON) {
    printSequenceState('p');
    
    playButtonPressed = false; // reset Play button state
  } else if (resetButtonPressed && !GREEN_ON) {
    printSequenceState('r');

    resetButtonPressed = false; // reset Reset button state
  } else {
    readPushButtons();

    // play the sound corresponding to the button pressed
    for (unsigned int i = 0; i < pushbuttonsCount; i++) {
      if (pushbutton_value[i] && debounceReady() && !RED_ON) {
        printButtonState(i);

        Serial.print("pressed ");        
        Serial.println(pushbutton_name[i]);
        
        // only play MAX_SEQUENCE sounds
        if (sequence_idx < MAX_SEQUENCE) {
          if (!GREEN_ON) {
            Serial.println("You have 18 seconds to play your best beat!");
            // start the timer which will let you play for 18 seconds
            startTimer();
          }

          play(i);
        }

        addToSequence(i);
        delay(200);
      }
    }
  }
}

// set the sound length and wholenote value for each sound
void setup_sounds()
{
  notes[0] = sizeof(sound1) / sizeof(sound1[0]) / 2;
  wholenote[0] = (60000 * 4) / tempo[0];

  notes[1] = sizeof(sound2) / sizeof(sound2[0]) / 2;
  wholenote[1] = (60000 * 4) / tempo[1];

  notes[2] = sizeof(sound3) / sizeof(sound3[0]) / 2;
  wholenote[2] = (60000 * 4) / tempo[2];

  notes[3] = sizeof(sound4) / sizeof(sound4[0]) / 2;
  wholenote[3] = (60000 * 4) / tempo[3];

  notes[4] = sizeof(sound5) / sizeof(sound5[0]) / 2;
  wholenote[4] = (60000 * 4) / tempo[4];

  notes[5] = sizeof(sound6) / sizeof(sound6[0]) / 2;
  wholenote[5] = (60000 * 4) / tempo[5];
}

// set interrupt for the Play (PD3) and Reset buttons (PD2)
void setup_interrupts(){
  cli();

  // BUTTONS INTERRUPTS

  // Set PD3 (INT1) and PD2 (INT0) as inputs with internal pull-up resistors
  DDRD &= ~((1 << PD3) | (1 << PD2));
  PORTD |= (1 << PD3) | (1 << PD2);

  // Configure INT1 (PD3) for falling edge triggered interrupt
  EICRA |= (1 << ISC11);
  EICRA &= ~(1 << ISC10);

  // Enable INT1 interrupt
  EIMSK |= (1 << INT1);

  // Configure INT0 (PD2) for falling edge triggered interrupt
  EICRA |= (1 << ISC01);
  EICRA &= ~(1 << ISC00);

  // Enable INT0 interrupt
  EIMSK |= (1 << INT0); 

  sei();
}

// initialize Timer1
void setup_timer1() {
  interval = 0;

  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  OCR1A = 31249;            // compare match register 16MHz/256/2Hz-1
  TCCR1B |= (1 << WGM12);   // CTC mode
  TCCR1B |= (1 << CS12);    // 256 prescaler 
}

// read the output voltage and update the button pressed with value 1 (true)
void readPushButtons()
{
  float VR_Calculated, RT, voltage_maximum, voltage_minimum;
  float V_measured = analogRead(inPin);  
  
  V_measured = V_measured * ((float)sourceVoltage / (pow(2, BIT_SOLUTION) - 1));

  for (unsigned int i = 0; i < pushbuttonsCount; i++) {
    RT = 1 / ((1 / ((i + 1) * (float)resistorValue) + (1 / (float)resistorToGround)));
  
    VR_Calculated = (sourceVoltage * RT) / ((pushbuttonsCount - (i + 1)) * (float)resistorValue + RT);

    if (i + 1 == pushbuttonsCount) {
      VR_Calculated = sourceVoltage;//for last push button, voltage is the same as power supply
    }

    voltage_minimum = VR_Calculated - V_tolerance * VR_Calculated;
    voltage_maximum = VR_Calculated + V_tolerance * VR_Calculated;

    if (V_measured <= voltage_maximum && V_measured >= voltage_minimum) {
        pushbutton_value[i] = true;
        delay(pushDelayTime);
    } else {
      pushbutton_value[i] = false;
    }
  }
}

// add a sound to the sequence array
void addToSequence(int i) {
  if (sequence_idx < MAX_SEQUENCE){
    sequence[sequence_idx] = i;
    sequence_idx++;
  } else {
    lcd.clear();         
    lcd.setCursor(0, 0);
    lcd.print("Max Size Reached");         
    lcd.setCursor(1, 1);
    lcd.print("Need To Reset!");
  }
}

// play only one sound at a time
void play(int i) 
{
  // Disable INT1 interrupt
  // Used for not triggering the INT intrerrupts
  EIMSK &= ~(1 << INT1);

  int *sound = all_sounds[i];
  int length = notes[i] * 2;

  for (int j = 0; j < length; j = j + 2) {
    int noteDuration = wholenote[i] / sound[j + 1];
    tone(buzzer, sound[j], noteDuration);
    int pauseBetweenNotes = noteDuration * 1.30;
    delay(pauseBetweenNotes);
    noTone(buzzer);
  }

  // Re-enable INT1 interrupt
  EIMSK |= (1 << INT1);
}

// play all the sounds in the sequence so far
void playSequence() {
  for (int i = 0; i < sequence_idx; i++) {
    play(sequence[i]);
  }
}

// used for printing and formating the value of the button pressed
void printButtonState(int i) 
{
  lcd.clear();         
  lcd.setCursor(1,0);
  lcd.print("You pressed ");
  lcd.print(pushbutton_name[i]);
  
  int pos;

  switch (i) {
    case 0:
      pos = 3;
      break;
    case 2:
      pos = 4;
      break;
    case 3:
      pos = 4;
      break;
    default:
      pos = 5;
      break;
  }
  
  lcd.setCursor(pos, 1);
  lcd.print(sounds_name[i]);
}

// used for printing and formating Play/Reset messages
void printSequenceState(char command) 
{
  switch (command) {
    case 'p':
      if (sequence_idx) {
        lcd.clear();         
        lcd.setCursor(0, 0);
        lcd.print("Playing Sequence");

        playSequence();
        delay(1000);

        lcd.clear();         
        lcd.setCursor(1, 0);
        lcd.print("Sequence Ended");         
        lcd.setCursor(1, 1);
        lcd.print("Resume Playing");
      } else {
        lcd.clear();         
        lcd.setCursor(3, 0);
        lcd.print("No Sequence");
        lcd.setCursor(4, 1);
        lcd.print("To play");

        delay(2000);

        lcd.clear();         
        lcd.setCursor(1, 0);
        lcd.print("Start Playing!");
      }
      break;
    case 'r':
      if (sequence_idx) {
        lcd.clear();         
        lcd.setCursor(4, 0);
        lcd.print("Reseting");         
        lcd.setCursor(4, 1);
        lcd.print("Sequence");

        memset(sequence, 0, MAX_SEQUENCE);
        sequence_idx = 0;
        delay(2000);

        lcd.clear();         
        lcd.setCursor(1, 0);
        lcd.print("Sequence Reset");         
        lcd.setCursor(1, 1);
        lcd.print("Start Playing");
      } else {
        lcd.clear();         
        lcd.setCursor(3, 0);
        lcd.print("No Sequence");
        lcd.setCursor(4, 1);
        lcd.print("To reset");

        delay(2000);

        lcd.clear();         
        lcd.setCursor(1, 0);
        lcd.print("Start Playing!");
      }
      break;
  }
}

// debounce the button so it only registers 1 press
bool debounceReady() 
{
  int buttonValue = digitalRead(inPin);
 
  current_time = millis();
  if (current_time - old_time > 1000 && buttonValue == 0) {
    old_time = current_time;

    return true;
  }

  return false;
}

// start the timer by turning the green LED on
void startTimer() {
  // Enable the Timer1 compare match A interrupt
  TIMSK1 |= (1 << OCIE1A);

  GREEN_ON = true;
  digitalWrite(GREEN_LED, HIGH);   // Turn off the LED
}